class AdqlColumn:
  def __init__(self, name):
    self._name = name
    self._table = None
    
  def __gt__(self, other):
    return AdqlColumn(f'({self._name} > {str(other)})')
  
  def __lt__(self, other):
    return AdqlColumn(f'({self._name} < {str(other)})')
  
  def __ge__(self, other):
    return AdqlColumn(f'({self._name} >= {str(other)})')
  
  def __le__(self, other):
    return AdqlColumn(f'({self._name} <= {str(other)})')
  
  def __eq__(self, other):
    return AdqlColumn(f'({self._name} = {str(other)})')
  
  def __ne__(self, other):
    return AdqlColumn(f'({self._name} != {str(other)})')
  
  def __or__(self, other):
    return AdqlColumn(f'({self._name} OR {str(other)})')
  
  def __and__(self, other):
    return AdqlColumn(f'({self._name} AND {str(other)})')
  
  def __add__(self, other):
    return AdqlColumn(f'({self._name} + {str(other)})')
  
  def __sub__(self, other):
    return AdqlColumn(f'({self._name} - {str(other)})')
  
  def __mul__(self, other):
    return AdqlColumn(f'({self._name} * {str(other)})')
  
  def __truediv__(self, other):
    return AdqlColumn(f'({self._name} / {str(other)})')
  
  def __mod__(self, other):
    return AdqlColumn(f'({self._name} % {str(other)})')
  
  def __pow__(self, other):
    return AdqlColumn(f'(POWER({self._name}, {str(other)}))')
  
  def __neg__(self):
    return AdqlColumn(f'(-{self._name})')
  
  def __pos__(self):
    return AdqlColumn(f'(+{self._name})')
  
  def __pos__(self):
    return AdqlColumn(f'(NOT({self._name}))')
  
  def between(self, first, second):
    return AdqlColumn(f'({self._name} BETWEEN {str(first)} AND {str(second)})')
  
  def like(self, other):
    return AdqlColumn(f"({self._name} LIKE '{str(other)}')")

  def __str__(self) -> str:
    return self._name
    
  def set_table(self, table):
    self._table = table
    
  def get_table(self):
    return self._table
  
  def get_name(self):
    return self._name
    
    
    
class AdqlTable:
  # def join(self, table: 'AdqlTable'):
  #   return AdqlTable()
  
  def __str__(self):
    return self._name
  
  def set_schema(self, schema):
    self._schema = schema
  
  def get_schema(self):
    return self._schema
  
  def get_name(self):
    return self._name
  
  def config(self):
    cols = [v for k, v in self.__class__.__dict__.items() if not k.startswith('_')]
    for col in cols:
      col.set_table(self)




class AdqlSchema:
  def get_name(self):
    return self._name
  
  def config(self):
    tables = [v for k, v in self.__class__.__dict__.items() if not k.startswith('_')]
    for table in tables:
      table.set_schema(self)
      table.config()
      
      

class AdqlDatabase:
  def config(self):
    schemas = [v for k, v in self.__class__.__dict__.items() if not k.startswith('_')]
    for schema in schemas:
      schema.config()


"""
Distance transformation
"""


from typing import Sequence

from astropy import units as u
from astropy.cosmology import FlatLambdaCDM, Planck18


def arcsec2mpc(
  distance: float | u.Quantity | Sequence[float | u.Quantity], 
  z: float | Sequence[float], 
  cosmology: FlatLambdaCDM = Planck18
):
  """
  Converts arcsecs to Mpc at a given redshift using the specified cosmology
  This function can be used to convert a single distance or multiple distances 
  with a single call.

  Parameters
  ----------
  distance: float or Quantity
    The distance in arcsec to be converted. This argument accepts a float variable 
    or an array-like object (python list, numpy array, pandas series) of floats.
    If a float or sequence of floats is passed, this function interprets it
    as a arcsec unit. If a quantity is passed, this funcion convert the given
    unit (e.g. arcmin or deg) to arcsec. Therefore, if a Quantity object is
    passed, there is no need to convert it.
  z: float or Sequence[float]
    The redshift. This argument can be a single redshift, so this function 
    considers that all distances are in the same redshift, or a sequence of
    redshift of same number of elements of the sequence passed in distance 
    argument, so this function considers that the i-th distance is at i-th
    redshift.
  cosmology: FlatLambdaCDM, optional.
    An astropy cosmology. This function uses Planck 2018 cosmology as default
    with parameters from Planck Collaboration (2020) Table 2 (TT, TE, EE
    + lowE + lensing + BAO) [#P18]_. That is also possible to use a custom
    cosmology, passing a FlatLambdaCDM instance or any cosmology realizations
    from astropy package [#AR]_.

  Returns
  -------
  Quantity
    The converted distance in Mpc
    
  References
  ----------
  .. [#P18] Planck Collaboration, et. al. (2020). Planck 2018 results. VI.
      Cosmological parameters. Astronomy \& Astrophysics, 641, A6.
      `<https://ui.adsabs.harvard.edu/abs/2020A%26A...641A...6P/abstract>`_
  .. [#AR] Included Cosmology Realizations - Astropy Documentation
      `<https://docs.astropy.org/en/stable/cosmology/realizations.html>`_
  """
  if isinstance(distance, u.Quantity):
    distance = distance.to(u.arcec)
  else:
    distance *= u.arcsec

  return cosmology.kpc_proper_per_arcmin(z).to(u.Mpc/u.arcsec)*(distance)


def mpc2arcsec(
  distance: float | u.Quantity | Sequence[float | u.Quantity], 
  z: float | Sequence[float], 
  cosmology: FlatLambdaCDM = Planck18
):
  """
  Converts Mpc to arcsecs at a given redshif using the specified cosmology
  This function can be used to convert a single distance or multiple distances 
  with a single call.

  Parameters
  ----------
  distance: float or Quantity
    The distance in Mpc to be converted. This argument accepts a float variable 
    or an array-like object (python list, numpy array, pandas series) of floats.
    If a float or sequence of floats is passed, this function interprets it
    as a Mpc unit. If a quantity is passed, this funcion convert the given
    unit to Mpc. Therefore, if a Quantity object is passed, there is no 
    need to convert it.
  z: float or Sequence[float]
    The redshift. This argument can be a single redshift, so this function 
    considers that all distances are in the same redshift, or a sequence of
    redshift of same number of elements of the sequence passed in distance 
    argument, so this function considers that the i-th distance is at i-th
    redshift.
  cosmology: FlatLambdaCDM, optional
    An astropy cosmology. This function uses Planck 2018 cosmology as default
    with parameters from Planck Collaboration (2020) Table 2 (TT, TE, EE
    + lowE + lensing + BAO) [#P18]_. That is also possible to use a custom
    cosmology, passing a FlatLambdaCDM instance or any cosmology realizations
    from astropy package [#AR]_.

  Returns
  -------
  Quantity
    The converted distance in arcsec
  
  References
  ----------
  .. [#P18] Planck Collaboration, et. al. (2020). Planck 2018 results. VI.
      Cosmological parameters. Astronomy \& Astrophysics, 641, A6.
      `<https://ui.adsabs.harvard.edu/abs/2020A%26A...641A...6P/abstract>`_
  .. [#AR] Included Cosmology Realizations - Astropy Documentation
      `<https://docs.astropy.org/en/stable/cosmology/realizations.html>`_
  """
  if isinstance(distance, u.Quantity):
    distance = distance.to(u.Mpc)
  else:
    distance *= u.Mpc
    
  return cosmology.arcsec_per_kpc_proper(z).to(u.arcsec/u.Mpc)*(distance)



if __name__ == '__main__':
  print(arcsec2mpc(2.3, 0.03))
  print(arcsec2mpc([2.3, 4.4], 0.03))
  print(arcsec2mpc([2.3, 4.4], [0.3, 0.03]))


"""
Input/Output operations
"""

import bz2
import concurrent.futures
import re
from io import BufferedIOBase, RawIOBase, StringIO, TextIOBase
from pathlib import Path
from typing import Any, BinaryIO, Callable, Dict, Literal, Sequence

import numpy as np
import pandas as pd
import requests
from astropy.io import fits, votable
from astropy.table import Table
from PIL import Image, ImageOps
from pypdf import PdfWriter
from tqdm import tqdm

RANDOM_SEED = 42
PathOrFile = str | Path | BufferedIOBase | RawIOBase | TextIOBase
TableLike = Table | pd.DataFrame

def read_table(
  path: TableLike | PathOrFile,
  fmt: str | None = None,
  columns: Sequence[str] | None = None,
  low_memory: bool = False,
  comment: str | None = None,
  na_values: Sequence[str] | Dict[str, Sequence[str]] = None,
  keep_default_na: bool = True,
  na_filter: bool = True,
  header: Literal['infer'] | int | Sequence[int] = 'infer',
  col_names: Sequence[str] | None = None,
) -> pd.DataFrame:
  """
  This function tries to detect the table type comparing the file extension and
  returns a pandas dataframe of the loaded table.
  
  Supported table types:
  
    =============== ===========================
    Table Type      Extensions
    =============== ===========================
    Fits            .fit, .fits, .fz
    Votable         .vo, .vot, .votable, .xml
    ASCII           .csv, .tsv, .dat
    Heasarc         .tdat
    Arrow           .parquet, .feather
    =============== ===========================

  Parameters
  ----------
  path : str or Path
    Path to the table to be read.
  
  fmt : str | None
    Specify the file format manually to avoid inference by file extension. This
    parameter can be used to force a specific parser for the given file.
  
  columns : Sequence[str] | None
    If specified, only the column names in list will be loaded. Can be used to
    reduce memory usage.
  
  low_memory : bool
    Internally process the file in chunks, resulting in lower memory use while 
    parsing, but possibly mixed type inference. To ensure no mixed types either 
    set False, or specify the type with the dtype parameter. Note that the 
    entire file is read into a single DataFrame regardless, use the chunksize 
    or iterator parameter to return the data in chunks. (Only valid with C parser).
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.
  
  comment : str | None
    Character indicating that the remainder of line should not be parsed. 
    If found at the beginning of a line, the line will be ignored altogether. 
    This parameter must be a single character. Like empty lines 
    (as long as ``skip_blank_lines=True``), fully commented lines are ignored 
    by the parameter header but not by skiprows. For example, if ``comment='#'``, 
    parsing ``#empty\\na,b,c\\n1,2,3`` with ``header=0`` will result in 
    ``'a,b,c'`` being treated as the header.
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.
  
  na_values: Hashable, Iterable of Hashable or dict of {HashableIterable}
    Additional strings to recognize as ``NA``/``NaN``. If ``dict`` passed, specific 
    per-column ``NA`` values. By default the following values are interpreted 
    as `NaN`: “ “, “#N/A”, “#N/A N/A”, “#NA”, “-1.#IND”, “-1.#QNAN”, “-NaN”, 
    “-nan”, “1.#IND”, “1.#QNAN”, “<NA>”, “N/A”, “NA”, “NULL”, “NaN”, “None”, 
    “n/a”, “nan”, “null “.
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.
  
  keep_default_na : bool 
    Whether or not to include the default ``NaN`` values when parsing the data. 
    Depending on whether ``na_values`` is passed in, the behavior is as follows:

    - If ``keep_default_na`` is ``True``, and ``na_values`` are specified, 
      `na_values` is appended to the default NaN values used for parsing.
    - If ``keep_default_na`` is ``True``, and ``na_values`` are not specified, only the 
      default ``NaN`` values are used for parsing.
    - If ``keep_default_na`` is ``False``, and ``na_values`` are specified, only 
      the ``NaN`` values specified na_values are used for parsing.
    - If ``keep_default_na`` is ``False``, and ``na_values`` are not specified, 
      no strings will be parsed as ``NaN``.

    Note that if ``na_filter`` is passed in as ``False``, the ``keep_default_na`` and 
    ``na_values`` parameters will be ignored.
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.
  
  na_filter : bool
    Detect missing value markers (empty strings and the value of ``na_values``). 
    In data without any ``NA`` values, passing ``na_filter=False`` can improve the 
    performance of reading a large file.
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.
      
  header : 'infer' or int or Sequence[int]
    Row number(s) containing column labels and marking the start of the data 
    (zero-indexed). Default behavior is to infer the column names: if no ``names``
    are passed the behavior is identical to ``header=0`` and column names are 
    inferred from the first line of the file, if column names are passed 
    explicitly to ``names`` then the behavior is identical to ``header=None``. 
    Explicitly pass ``header=0`` to be able to replace existing names. The 
    header can be a list of integers that specify row locations for a 
    `pandas.MultiIndex` on the columns e.g. ``[0, 1, 3]``. Intervening rows 
    that are not specified will be skipped (e.g. 2 in this example is skipped). 
    Note that this parameter ignores commented lines and empty lines if 
    ``skip_blank_lines=True``, so ``header=0`` denotes the first line of data 
    rather than the first line of the file.
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.
    
  col_names : Sequence[str]
    Sequence of column labels to apply. If the file contains a header row, 
    then you should explicitly pass ``header=0`` to override the column names. 
    Duplicates in this list are not allowed.
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.

  Notes
  -----
  The Transportable Database Aggregate Table (TDAT) type is a data structure 
  created by NASA's Heasarc project and a very simple parser was implemented
  in this function due to lack of support in packages like pandas and astropy. 
  For more information, see [#TDAT]_

  Returns
  -------
  pd.DataFrame
    The table as a pandas dataframe

  Raises
  ------
  ValueError
    Raises an error if the file extension can not be detected
    
  References
  ----------
  .. [#TDAT] Transportable Database Aggregate Table (TDAT) Format.
      `<https://heasarc.gsfc.nasa.gov/docs/software/dbdocs/tdat.html>`_
  """
  is_file_like = False
  if isinstance(path, (str, Path)):
    path = Path(path)
    fmt = fmt or path.suffix
  elif isinstance(path, pd.DataFrame):
    return path
  elif isinstance(path, Table):
    df = path.to_pandas()
    if columns:
      df = df[columns]
    return df
  elif isinstance(path, (RawIOBase, BufferedIOBase, TextIOBase)):
    is_file_like = True
  
  if fmt.startswith('.'):
    fmt = fmt[1:]

  if fmt in ('fit', 'fits', 'fz'):
    with fits.open(path) as hdul:
      table_data = hdul[1].data
      table = Table(data=table_data)
      if columns is None:
        columns = [name for name in table.colnames if len(table[name].shape) <= 1]
      df = table[columns].to_pandas()
      return df
  elif fmt in ('dat', 'tsv'):
    optional_params = {}
    if col_names:
      optional_params = {'names': col_names}
    return pd.read_csv(
      path, 
      sep='\s+', 
      usecols=columns, 
      low_memory=low_memory,
      comment=comment,
      na_values=na_values,
      keep_default_na=keep_default_na,
      na_filter=na_filter,
      header=header,
      **optional_params,
    )
  elif fmt == 'csv':
    optional_params = {}
    if col_names:
      optional_params = {'names': col_names}
    return pd.read_csv(
      path, 
      usecols=columns, 
      low_memory=low_memory,
      comment=comment,
      na_values=na_values,
      keep_default_na=keep_default_na,
      na_filter=na_filter,
      header=header,
      **optional_params,
    )
  elif fmt == 'parquet':
    return pd.read_parquet(
      path, 
      columns=columns,
    )
  elif fmt == 'feather':
    return pd.read_feather(
      path, 
      columns=columns
    )
  elif fmt == 'tdat':
    if is_file_like: raise ValueError('Not implemented for file-like objects')
    path = Path(path)
    content = path.read_text()
    header = re.findall(r'line\[1\] = (.*)', content)[0].replace(' ', '|')
    data = content.split('<DATA>\n')[-1].split('<END>')[0].replace('|\n', '\n')
    tb = header + '\n' + data
    return pd.read_csv(
      StringIO(tb), 
      sep='|', 
      usecols=columns, 
      low_memory=low_memory
    )
  elif fmt in ('vo', 'vot', 'votable', 'xml'):
    result = votable.parse_single_table(path)
    table = result.to_table(use_names_over_ids=True)
    # table = result.get_first_table().to_table(use_names_over_ids=True)
    df = table.to_pandas()
    if columns:
      df = df[columns]
    return df

  raise ValueError(
    'Can not infer the load function for this table based on suffix. '
    'Please, use a specific loader.'
  )


def write_table(data: TableLike, path: PathOrFile, fmt: str | None = None):
  if isinstance(path, str):
    fmt = fmt or Path(path).suffix
    Path(path).parent.mkdir(parents=True, exist_ok=True)
  elif isinstance(path, Path):
    fmt = fmt or path.suffix
    path.parent.mkdir(parents=True, exist_ok=True)
    path = str(path.absolute())
  
  if isinstance(data, Table):
    data = data.to_pandas()
  
  if fmt.startswith('.'):
    fmt = fmt[1:]
  
  if fmt in ('fit', 'fits'):
    Table.from_pandas(data).write(path, overwrite=True, format='fits')
  elif fmt == 'csv':
    data.to_csv(path, index=False)
  elif fmt == 'parquet':
    data.to_parquet(path, index=False)
  elif fmt == 'dat':
    data.to_csv(path, index=False, sep=' ')
  elif fmt == 'tsv':
    data.to_csv(path, index=False, sep='\t')
  elif fmt == 'html':
    data.to_html(path, index=False)
  elif fmt == 'feather':
    data.to_feather(path, index=False)
  elif fmt in ('vo', 'vot', 'votable', 'xml'):
    t = Table.from_pandas(data)
    votable.writeto(t, path)




def read_image(path: str | Path) -> np.ndarray:
  """
  Load image from local storage to numpy array.
  Supports several types of files, incluing ``.jpg``, ``.png``, ``.npy``,
  ``.npz``, ``.fits``

  Parameters
  ----------
  path: str or pathlib.Path
    Path to the desired file

  Returns
  -------
  numpy.ndarray
    Image converted (if needed) to a numpy array.
  """
  path = Path(path)
  if path.suffix in ('.jpg', '.jpeg', '.png'):
    img = Image.open(path)
    return np.asarray(img)
  elif path.suffix in ('.npy', '.npz'):
    return np.load(path)
  elif path.suffix in ('.fits', '.fit', '.fz'):
    im = fits.getdata(path)
    if np.argmin(im.shape) == 0:
      # convert from (c, h, w) to (h, w, c)
      im = np.moveaxis(im, 0 , -1)
    return im
  
  

def write_image(data: np.ndarray, path: str | Path):
  path = Path(path)
  path.parent.mkdir(parents=True, exist_ok=True)

  if path.suffix in ('.jpg', '.png'):
    ImageOps.flip(Image.fromarray(data, 'RGB')).save(path)
  elif path.suffix in ('.fits', '.fit'):
    fits.ImageHDU(np.moveaxis(data, -1, 0)).writeto(path, overwrite=True)
  elif path.suffix == '.npy':
    np.save(path, data)
  elif path.suffix == '.npz':
    np.savez(path, *data)
    
    

def compress_fits(
  file: PathOrFile,
  compress_type: str = 'HCOMPRESS_1',
  hcomp_scale: int = 3,
  quantize_level: int = 10,
  quantize_method: int = -1,
  ext: int = 0,
  save_path: str | Path = None,
  replace: bool = True,
):
  hdul = fits.open(file)

  if ext >= len(hdul):
    raise ValueError(f'Trying to access ext {ext}, max ext is {len(hdul)-1}')

  if save_path.exists() and not replace:
    return None

  comp = None

  try:
    comp = fits.CompImageHDU(
      data=hdul[ext].data,
      header=hdul[ext].header,
      compression_type=compress_type,
      hcomp_scale=hcomp_scale,
      quantize_level=quantize_level,
      quantize_method=quantize_method,
      dither_seed=RANDOM_SEED,
    )
    if save_path:
      comp.writeto(
        save_path,
        overwrite=replace,
      )
  except OSError as e:
    pass

  return comp



def download_file(
  url: str,
  save_path: str | Path,
  replace: bool = False,
  http_client: requests.Session = None,
  extract: bool = False,
  return_bytes: bool = False,
) -> bytes | None:
  save_path = Path(save_path)

  if not replace and save_path.exists():
    return None

  http_client = http_client or requests

  if not save_path.parent.exists():
    save_path.parent.mkdir(parents=True, exist_ok=True)

  r = http_client.get(url, allow_redirects=True)

  if r.status_code == 200:
    if extract:
      file_bytes = bz2.decompress(r.content)
    else:
      file_bytes = r.content

    if return_bytes:
      return file_bytes

    with open(str(save_path.resolve()), 'wb') as f:
      f.write(file_bytes)

  return None


def parallel_function_executor(
  func: Callable,
  params: Sequence[Dict[str, Any]] = [],
  workers: int = 2,
  unit: str = 'it'
):
  with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
    futures = []

    for i in range(len(params)):
      futures.append(executor.submit(
        func,
        **params[i]
      ))

    for future in tqdm(
      concurrent.futures.as_completed(futures),
      total=len(futures),
      unit=unit
    ):
      try:
        future.result()
      except Exception as e:
        pass


def batch_download_file(
  urls: Sequence[str],
  save_path: Sequence[str] | Sequence[Path],
  replace: bool = False,
  http_client: requests.Session = None,
  workers: int = 2,
):
  params = [
    {
      'url': _url,
      'save_path': Path(_save_path),
      'replace': replace,
      'http_client': http_client
    }
    for _url, _save_path in zip(urls, save_path)
  ]
  parallel_function_executor(download_file, params, workers=workers, unit='file')




def merge_pdf(pdfs: Sequence[str | Path], output_path: str | Path):
  """
  Merge a sequence of pdf files into one single file

  Parameters
  ----------
  pdfs : Sequence[str  |  Path]
    Input paths, must be a sequence
  output_path : str | Path
    The output file path
  """
  merger = PdfWriter()
  for pdf in pdfs:
    p = Path(pdf)
    if p.exists():
      merger.append(p)
  merger.write(output_path)
  merger.close()
  
  
  
  


if __name__ == '__main__':
  df = pd.DataFrame({'RA': [193.20100, 193.31000], 'DEC': [-15.43320, -15.38000]})
  write_table(df, 'test.xml')



import concurrent.futures
import secrets
import tempfile
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Sequence
from urllib.parse import urlencode

import pandas as pd
import requests
import tqdm

from astromodule.io import (PathOrFile, batch_download_file, download_file,
                            write_table)
from astromodule.utils import append_query_params


class ImagingService(ABC):
  def __init__(self, image_format):
    self.image_format = image_format


  @abstractmethod
  def cutout(self, ra: float, dec: float, save_path: Path):
    pass


  @abstractmethod
  def batch_cutout(self, ra: List[float], dec: List[float], save_path: List[Path]):
    pass


  def _download_rgb(
    self,
    ra: float,
    dec: float,
    save_path: Path,
    width: int = 256,
    height: int = 256,
    scale: float = 0.55,
    opt: str = ''
  ):
    image_url = append_query_params(self.imaging_url, {
      'ra': ra,
      'dec': dec,
      'width': width,
      'height': height,
      'scale': scale,
      'opt': opt
    })
    download_file(image_url, save_path)


  def _batch_download_rgb(
    self,
    ra: Sequence[float],
    dec: Sequence[float],
    save_path: Sequence[Path],
    workers: int = None,
    **kwargs
  ):
    with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
      futures = []

      for i in range(len(ra)):
        futures.append(executor.submit(
          self.download_rgb,
          ra=ra[i],
          dec=dec[i],
          save_path=save_path[i],
          **kwargs
        ))

      for _ in tqdm.tqdm(
        concurrent.futures.as_completed(futures),
        total=len(futures),
        unit=' file'
      ):
        pass





class TapService:
  def __init__(self, url):
    self.url = url
    self.http_client = requests.Session()

  def sync_query(self, query: str, save_path: PathOrFile):
    params = {
      'request': 'doQuery',
      'version': 1.0,
      'lang': 'ADQL',
      'phase': 'run',
      'format': 'csv',
      'query': query
    }
    req_url = self.url + '?' + urlencode(params)
    save_path = Path(save_path)
    save_path.parent.mkdir(parents=True, exist_ok=True)
    download_file(req_url, save_path, replace=True, http_client=self.http_client)

  def batch_sync_query(
    self,
    queries: Sequence[str],
    save_paths: PathOrFile | Sequence[PathOrFile],
    join_outputs: bool = False,
    workers: int = 3
  ):
    params = {
      'request': 'doQuery',
      'version': 1.0,
      'lang': 'ADQL',
      'phase': 'run',
      'format': 'csv'
    }
    urls = [self.url + '?' + urlencode({**params, 'query': q}) for q in queries]

    save_paths_aux = save_paths
    if join_outputs:
      tmp_folder = Path(tempfile.gettempdir()) / f'tap_{secrets.token_hex(3)}'
      tmp_folder.mkdir(parents=True)
      save_paths_aux = [tmp_folder / f'{i}.csv' for i in range(len(queries))]

    batch_download_file(
      urls,
      save_path=save_paths_aux,
      http_client=self.http_client,
      workers=workers
    )

    if join_outputs:
      combined_csv = pd.concat([
        pd.read_csv(f, comment='#') for f in tmp_folder.glob('*.csv')
      ])
      save_paths = Path(save_paths)
      save_paths.parent.mkdir(parents=True, exist_ok=True)
      write_table(combined_csv, save_paths)


import os
import subprocess
from argparse import ArgumentParser
from time import sleep


def cbpf_up(args):
  user = os.environ.get('CBPF_USER')
  password = os.environ.get('CBPF_PASS')
  local = args.local[0]
  remote = args.remote[0]
  cmd = f"sshpass -p {password} rsync --mkpath -r -v --progress -e 'ssh -p 13900' '{local}' {user}@tiomno.cbpf.br:'{remote}'"
  subprocess.call(cmd, shell=True)
  

def cbpf_down(args):
  user = os.environ.get('CBPF_USER')
  password = os.environ.get('CBPF_PASS')
  local = args.local[0]
  remote = args.remote[0]
  cmd = f"sshpass -p {password} rsync --mkpath -r -v --progress -e 'ssh -p 13900' {user}@tiomno.cbpf.br:'{remote}' '{local}'"
  it = 0
  while it < args.repeat:
    print(f'Execution {it} of {args.repeat}')
    subprocess.call(cmd, shell=True)
    sleep(args.delay)
    it += 1
  

def cbpf_ssh(args):
  user = os.environ.get('CBPF_USER')
  password = os.environ.get('CBPF_PASS') 
  cmd = f"sshpass -p {password} ssh -p 13900 {user}@tiomno.cbpf.br"
  subprocess.call(cmd, shell=True)


def cbpf():
  parser = ArgumentParser(
    prog='cbpf', 
    description='CBPF cluster access'
  )
  
  subparser = parser.add_subparsers(dest='subprog')
  
  down = subparser.add_parser('down')
  down.add_argument('-r', '--repeat', type=int, default=1, action='store', help='number of times to repeat')
  down.add_argument('-d', '--delay', type=int, default=120, action='store', help='delay time in seconds')
  down.add_argument('remote', nargs=1)
  down.add_argument('local', nargs='+')
  
  up = subparser.add_parser('up')
  up.add_argument('local', nargs=1) 
  up.add_argument('remote', nargs='+')
  
  subparser.add_parser('ssh')
  
  args = parser.parse_args()
  
  cmds = {
    'down': cbpf_down,
    'up': cbpf_up,
    'ssh': cbpf_ssh
  }
  
  handler = cmds.get(args.subprog)
  if handler:
    handler(args)
  else:
    parser.print_help()
  
  
  
  


def teiu_up(args):
  user = os.environ.get('TEIU_USER')
  password = os.environ.get('TEIU_PASS')
  local = args.local[0]
  remote = args.remote[0]
  cmd = f"sshpass -p {password} rsync --mkpath -r -v --progress -e ssh '{local}' {user}@teiu.iag.usp.br:'{remote}'"
  subprocess.call(cmd, shell=True)


def teiu_down(args):
  user = os.environ.get('TEIU_USER')
  password = os.environ.get('TEIU_PASS')
  local = args.local[0]
  remote = args.remote[0]
  cmd = f"sshpass -p {password} rsync --mkpath -r -v --progress -e ssh {user}@teiu.iag.usp.br:'{remote}' '{local}'"
  subprocess.call(cmd, shell=True)
  
  
def teiu_ssh(args):
  user = os.environ.get('TEIU_USER')
  password = os.environ.get('TEIU_PASS') 
  cmd = f"sshpass -p {password} ssh {user}@teiu.iag.usp.br"
  subprocess.call(cmd, shell=True)


def teiu():
  parser = ArgumentParser(
    prog='cbpf', 
    description='Teiu cluster access'
  )
  
  subparser = parser.add_subparsers(dest='subprog')
  
  down = subparser.add_parser('down')
  down.add_argument('remote', nargs=1)
  down.add_argument('local', nargs='+')
  
  up = subparser.add_parser('up')
  up.add_argument('local', nargs=1) 
  up.add_argument('remote', nargs='+')
  
  subparser.add_parser('ssh')
  
  args = parser.parse_args()
  
  cmds = {
    'down': teiu_down,
    'up': teiu_up,
    'ssh': teiu_ssh
  }
  
  handler = cmds.get(args.subprog)
  if handler:
    handler(args)
  else:
    parser.print_help()




def iguana_up(args):
  user = os.environ.get('IGUANA_USER')
  password = os.environ.get('IGUANA_PASS')
  local = args.local[0]
  remote = args.remote[0]
  cmd = f"sshpass -p {password} rsync --mkpath -r -v --progress -e ssh '{local}' {user}@iguana.iag.usp.br:'{remote}'"
  subprocess.call(cmd, shell=True)


def iguana_down(args):
  user = os.environ.get('IGUANA_USER')
  password = os.environ.get('IGUANA_PASS')
  local = args.local[0]
  remote = args.remote[0]
  cmd = f"sshpass -p {password} rsync --mkpath -r -v --progress -e ssh {user}@iguana.iag.usp.br:'{remote}' '{local}'"
  subprocess.call(cmd, shell=True)
  
  
def iguana_ssh(args):
  user = os.environ.get('IGUANA_USER')
  password = os.environ.get('IGUANA_PASS') 
  cmd = f"sshpass -p {password} ssh {user}@iguana.iag.usp.br"
  subprocess.call(cmd, shell=True)



def iguana():
  parser = ArgumentParser(
    prog='cbpf', 
    description='Iguana cluster access'
  )
  
  subparser = parser.add_subparsers(dest='subprog')
  
  down = subparser.add_parser('down')
  down.add_argument('remote', nargs=1)
  down.add_argument('local', nargs='+')
  
  up = subparser.add_parser('up')
  up.add_argument('local', nargs=1) 
  up.add_argument('remote', nargs='+')
  
  subparser.add_parser('ssh')
  
  args = parser.parse_args()
  
  cmds = {
    'down': iguana_down,
    'up': iguana_up,
    'ssh': iguana_ssh
  }
  
  handler = cmds.get(args.subprog)
  if handler:
    handler(args)
  else:
    parser.print_help()
  
  
if __name__ == "__main__":
  cbpf()


import unittest

import pandas as pd
from astropy import units as u

from astromodule.table import crossmatch


class TestCrossMatch(unittest.TestCase):
  def test_valid_data_expected_results(self):
    """
    Given two tables with valid data, it should return a pandas DataFrame 
    with the expected crossmatch results.
    """
    # Initialize the input tables
    table1 = pd.DataFrame({
      'RA': [10.0, 20.0, 30.0],
      'DEC': [45.0, 50.0, 55.0],
      'Name': ['Star1', 'Star2', 'Star3']
    })

    table2 = pd.DataFrame({
      'RA': [15.0, 25.0, 35.0],
      'DEC': [47.0, 52.0, 57.0],
      'Type': ['Galaxy', 'Galaxy', 'Galaxy']
    })

    # Invoke the stilts_crossmatch function
    result = crossmatch(table1, table2, ra1='RA', dec1='DEC', ra2='RA', dec2='DEC', radius=0)

    # Define the expected result
    expected_result = pd.DataFrame()
    
    print(result)
    print(expected_result)

    # Assert the result is equal to the expected result
    pd.testing.assert_frame_equal(result, expected_result)


"""
Simple implementation of a linear pipeline with multiprocessing support
"""

import inspect
import tempfile
from abc import ABC, abstractmethod
from copy import deepcopy
from pathlib import Path
from secrets import token_hex
from typing import Any, Dict, Sequence

import graphviz

from astromodule.io import parallel_function_executor
from astromodule.utils import SingletonMeta, filter_dict

# class PipelineStorage:
#   """
#   The storage used share resources between pipeline stage
  
#   Attributes
#   ----------
#   storage: dict
#     The data shared across pipeline stages
#   """
#   def __init__(self):
#     self._storage = {}
    
#   def write(self, key: str, value: Any):
#     """
#     Stores ``value`` identified by a ``key``

#     Parameters
#     ----------
#     key : str
#       The data identifier
#     value : Any
#       The data value
#     """
#     self._storage[key] = value
    
#   def write_many(self, **data):
#     """
#     Stores ``value`` identified by a ``key`` in ``data``

#     Parameters
#     ----------
#     **data: Any
#       The data to be stored
#     """
#     for key, value in data.items():
#       self._storage[key] = value
    
#   def read(self, key: str) -> Any:
#     """
#     Retrieve the ``value`` identified by a ``key``

#     Parameters
#     ----------
#     key : str
#       The data identifier
#     """
#     return self._storage.get(key)
    
#   def read_many(self, keys: Sequence[str]) -> Dict[str, Any]:
#     """
#     Retrieve the ``values`` identified by ``keys``

#     Parameters
#     ----------
#     keys : Sequence[str]
#       The data identifiers
#     """
#     return filter_dict(self._storage, keys)
  

class NotFoundKey:
  pass

  
class PipelineStorage(metaclass=SingletonMeta):
  """
  The storage used share resources between pipeline stage
  
  Attributes
  ----------
  storage: dict
    The data shared across pipeline stages
  """
  def __init__(self):
    self._storage = {}
    
  def write(self, key: str, value: Any, storage_id: str = 'shared'):
    """
    Stores ``value`` identified by a ``key``

    Parameters
    ----------
    key : str
      The data identifier
    value : Any
      The data value
    """
    if storage_id not in self._storage:
      self._storage[storage_id] = {}
    self._storage[storage_id][key] = value
    
  def write_many(self, storage_id: str = 'shared', **data):
    """
    Stores ``value`` identified by a ``key`` in ``data``

    Parameters
    ----------
    **data: Any
      The data to be stored
    """
    for key, value in data.items():
      self.write(key, value, storage_id)
    
  def read(self, key: str, storage_id: str = 'shared') -> Any:
    """
    Retrieve the ``value`` identified by a ``key``

    Parameters
    ----------
    key : str
      The data identifier
    """
    return self._storage.get(storage_id, {}).get(key)
    
  def read_many(self, keys: Sequence[str], storage_id: str = 'shared') -> Dict[str, Any]:
    """
    Retrieve the ``values`` identified by ``keys``

    Parameters
    ----------
    keys : Sequence[str]
      The data identifiers
    """
    return filter_dict(self._storage.get(storage_id, {}), keys)

  def del_storage(self, storage_id: str):
    """
    Delete storage defined by ``storage_id``

    Parameters
    ----------
    storage_id : str
      Storage identifier
    """
    if storage_id in self._storage:
      del self._storage[storage_id]
      
  def has_key(self, key: str, storage_id: str = 'shared'):
    return key in self._storage.get(storage_id, {})
  


class PipelineStage(ABC):
  """
  Base class for all pipeline stages
  
  Attributes
  ----------
  storage: PipelineStorage
    The pipeline storage. All resources of the pipeline are shared between 
    stages using this object. The `Pipeline` class creates this object
    during instantiation
    
  See Also
  --------
  astromodule.pipeline.Pipeline
  """
  requirements: Sequence[str] = []
  """
  A list of all resorces required by this pipeline stage. It's includes all
  shared data required by this stage using `get_data` or `run` arguments
  """
  products: Sequence[str] = []
  """
  A list of all resorces produced by this pipeline stage. It's includes all
  shared data produced by this stage using `set_data` 
  that can be accessed by another via `get_data` 
  or `run` arguments
  """
  
  @abstractmethod
  def run(self):
    """
    All concrete class of `PipelineStage` must implement this method.
    This method is called by `Pipeline.run` and `Pipeline.map_run` when
    executing the pipeline.
    
    .. note ::
      This method receives all needed data via
    
    See also
    --------
    astromodule.pipeline.Pipeline.run
    astromodule.pipeline.Pipeline.map_run
    """
    pass
  
  # @property
  # def storage(self) -> PipelineStorage:
  #   """
  #   The pipeline shared storage object

  #   Returns
  #   -------
  #   PipelineStorage
  #     The pipeline storage
  #   """
  #   return self._storage
  
  # @storage.setter
  # def storage(self, pipe_storage: PipelineStorage):
  #   """
  #   The pipeline shared storage object

  #   Parameters
  #   ----------
  #   pipe_storage : PipelineStorage
  #     A pipeline storage object
  #   """
  #   self._storage = pipe_storage
  
  @property
  def storage_id(self) -> str:
    return self._storage_id
  
  @storage_id.setter
  def storage_id(self, value: str):
    self._storage_id = value
  
  @property
  def name(self):
    return self.__class__.__name__
  
  def set_data(self, key: str, value: Any):
    """
    Stores shared data ``value`` identified by a ``key``

    Parameters
    ----------
    key : str
      The data identifier
    
    value : Any
      The data value
    """
    # self.storage.write(key, value)
    PipelineStorage().write(key, value, self.storage_id)
    
  def get_data(self, key: str) -> Path:
    """
    Retrieve shared data identified by a ``key``

    Parameters
    ----------
    key : str
      The data identifier
    """
    # return self.storage.read(key)
    storage = PipelineStorage()
    if storage.has_key(key, self.storage_id):
      return storage.read(key, self.storage_id)
    return storage.read(key, 'shared')



class Pipeline:
  """
  A simple linear pipeline implementation with mapping and multiprocessing
  support.
  
  Parameters
  ----------
  *stages: PipelineStage
    The stages of the pipeline
  
  verbose: bool, optional
    The verbosity flag, by default True.
    
  Attributes
  ----------
  storage: PipelineStorage
    The object used to share resources between all pipeline stages
    
  See Also
  --------
  astromodule.pipeline.PipelineStorage
  astromodule.pipeline.PipelineStage
  """
  def __init__(
    self, 
    *stages: PipelineStage, 
    verbose: bool = True, 
    req_list: Sequence[str] = None,
  ):
    self.verbose = verbose
    # self.storage = PipelineStorage()
    self.stages = [deepcopy(s) for s in stages]
    self.storage_id = token_hex(8)
    
    for stage in self.stages:
      stage.storage_id = self.storage_id
    
    if not req_list:
      req_list = self.get_stages_requirements()
    self.set_stages_requirements(req_list)
    self._req_list = req_list
  
  
  def set_stages_requirements(self, requirements: Sequence[str]):
    """
    Sets ``requirements`` attribute for all pipeline stages

    Parameters
    ----------
    requirements : Sequence[str]
      A list-like object of same size as pipeline stages. Each element of 
      this list is another list of required resources keys.
    """
    for stage, req in zip(self.stages, requirements):
      stage.requirements = req
      
      
  def get_stages_requirements(self) -> Sequence[str]:
    """
    Use code inspection to find required resources based in the assignature
    of `PipelineStage.run` method.

    Returns
    -------
    Sequence[str]
      A list of resources keys for each pipeline stage.
      
    See Also
    --------
    astromodule.pipeline.PipelineStage.run
    """
    all_reqs = []
    for stage in self.stages:
      reqs = list(inspect.signature(stage.run).parameters.keys())
      all_reqs.append(reqs)
    return all_reqs
  
    
  def run(self, validate: bool = True):
    """
    Validates and executes all stages of the pipeline

    Parameters
    ----------
    validate : bool, optional
      If `True`, a pipeline requirements validation will be performed using
      `validate` method. If `False`, the validation will be skiped, 
      by default True
    
    See Also
    --------
    astromodule.pipeline.Pipeline.validate
    """
    if validate and not self.validate():
      if self.verbose:
        print('Aborting pipeline execution due to validation fail')
      return 
    
    for i, stage in enumerate(self.stages, 1):
      if self.verbose:
        print(f'[{i} / {len(self.stages)}] {stage.name}')
      
      # kwargs = PipelineStorage().read_many(stage.requirements, self.storage_id)
      kwargs = {req: stage.get_data(req) for req in stage.requirements}
      outputs = stage.run(**kwargs)
      if isinstance(outputs, dict):
        for k, v in outputs.items():
          stage.set_data(k, v)
        # PipelineStorage().write_many(self.storage_id, **outputs)
      
      if self.verbose:
        print()
        
        
  def map_run(
    self, 
    key: str, 
    array: Sequence[Any], 
    workers: int = 2, 
    validate: bool = True
  ):
    """
    Validates and executes all pipeline steps in a similar way to the `run` 
    method, but using multiprocessing.
    This method has a similar implementation to MapReduce [#MapReduce]_, 
    in which a function is applied to all elements of a vector.
    In this case, the function is the pipeline itself and the vector 
    is specified by the ``array`` parameter.
    Thus, the pipeline is executed ``n`` times, where ``n`` is the size of 
    the ``array`` vector.
    For each execution, the pipeline creates an storage output with 
    identifier ``key`` whose value is the element of the vector that can
    be accessed via `PipelineStage.get_output`.

    Parameters
    ----------
    key : str
      The identifier for an element of ``array`` that can
      be accessed by a pipeline stage using `PipelineStage.get_output`.
    
    array : Sequence[Any]
      The data that will be mapped to pipeline
    
    workers : int, optional
      The number of parallel proccesses that will be spawned, by default 2
    
    validate : bool, optional
      If `True`, a pipeline requirements validation will be performed using
      `validate` method. If `False`, the validation will be skiped, 
      by default True
        
    See also
    --------
    astromodule.pipeline.Pipeline.run
    astromodule.pipeline.Pipeline.validate
    astromodule.pipeline.PipelineStage.get_output
    
    References
    ----------
    .. [#MapReduce] MapReduce - Wikipedia
      `<https://en.wikipedia.org/wiki/MapReduce>`_
    """
    if validate and not self.validate(ignore=[key]):
      if self.verbose:
        print('Aborting pipeline execution due to validation fail')
      return 
    
    if workers > 1:
      params = [{'key': key, 'data': d} for d in array]
      parallel_function_executor(
        func=self._pipe_executor, 
        params=params, 
        workers=workers, 
        unit='jobs'
      )
    else:
      for i, data in enumerate(array):
        print(f'[{i+1} / {len(array)}] Pipeline Start')
        self._pipe_executor(key=key, data=data, verbose=True)
        print()
      
      
  def validate(self, ignore: Sequence[str] = []) -> bool:
    """
    Validates the pipeline by checking whether all requirements for all 
    stages are satisfied
    
    Parameters
    ----------
    ignore : Sequence[str], optional
      Keys to ignore, by default []

    Returns
    -------
    bool
      `True` if all stages can retrieve the required resources (outputs
      and artifacts) correctly, `False` otherwise.
    """
    all_resources = set(ignore)
    problems = []
    for i, stage in enumerate(self.stages, 1):
      missing_req = set(stage.requirements) - all_resources
      if len(missing_req) > 0:
        problems.append({
          'stage_index': i, 
          'stage_name': stage.name, 
          'missing_req': missing_req
        })
      all_resources = all_resources.union(stage.products)
      
    if len(problems) > 0:
      print('Missing requirements:')
      for problem in problems:
        print(f'\t{problem["stage_index"]}. {problem["stage_name"]}')
        print(*[f'\t\t- {r}' for r in problem['missing_req']], sep='\n')
      return False
    return True
  
  
  def plot(self):
    """
    Plot the pipeline graph

    Returns
    -------
    graphviz
      A graphviz object containig the pipeline digraph that can be displayed
      in Jupyter Notebook
    """
    dot = graphviz.Digraph('Pipeline')
    for i, stage in enumerate(self.stages, 1):
      dot.node(str(i), f'{i}. {stage.name}')
    for i in range(1, len(self.stages)):
      dot.edge(str(i), str(i+1))
    dot.view(directory=tempfile.gettempdir(), cleanup=True)
    return dot
      
      
  def __repr__(self) -> str:
    """
    String object representation

    Returns
    -------
    str
      The object representation
    """
    p = [f'{i}. {s.name}' for i, s in enumerate(self.stages, 1)]
    p = '\n'.join(p)
    p = f'Pipeline:\n{p}'
    return p
  
  
  def __add__(self, other: PipelineStage) -> 'Pipeline':
    """
    Overcharge of the ``+`` operator that implements concatenation of
    pipelines or concatenation of a pipeline and a stage

    Parameters
    ----------
    other : Pipeline or PipelineStage
      The pipeline or stage that will be concatenated

    Returns
    -------
    Pipeline
      The resultant pipeline
    """
    if isinstance(other, PipelineStage):
      return Pipeline(*self.stages, other)
    elif isinstance(other, Pipeline):
      return Pipeline(*self.stages, *other.stages)
    
    
  def _pipe_executor(self, key: str, data: Any, verbose: bool = False):
    """
    Wrapper function that create a new pipeline instance and execute it to
    ensure isolation of different pipelines in parallel execution

    Parameters
    ----------
    key : str
      The resource name (identifier) that will be accessible to pipeline stages
    data : Any
      The value of the mapped resource
    """
    pipe = Pipeline(*self.stages, verbose=verbose, req_list=self._req_list)
    # pipe.storage.write(key, data)
    PipelineStorage().write(key, data)
    pipe.run(validate=False)
    PipelineStorage().del_storage(self.storage_id)
    del pipe


  
if __name__ == '__main__':
  import random
  import time
  class Stage1(PipelineStage):
    name = 'Stage 1'
    products = ['frame']
    
    def run(self, pipe = None):
      time.sleep(random.random())
      return {
        'frame': pipe * 2
      }
    
  class Stage2(PipelineStage):
    name = 'Stage 2'
    
    def run(self, frame):
      print(frame)
      time.sleep(random.random())
    
  p = Pipeline(Stage1(), Stage2())
  # p.plot()
  # print(p)
  # p.run()
  # p.validate()
  p.map_run('pipe', [1, 2, 3, 4, 5, 6], workers=2)


"""
Plotting functions
"""
from pathlib import Path
from typing import Sequence

import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import ImageGrid

from astromodule.io import read_image


def simple_image_axis(image: str | Path, ax = None):
  if not ax:
    fig, ax = plt.subplots()
  
  img = read_image(image)
  ax.imshow(img)
  ax.axis('off')
  return ax


def plot_axis_grid(
  images: Sequence[str | Path],
  cols: int, 
  rows: int, 
  width: float = 2.5, 
  height: float = 2.5, 
  size: float = None,
  title: str = None,
  show: bool = False,
  save_path: str | Path = None,
  axes_pad: float = 0.05,
):
  if size:
    width, height = size, size
  fig, axs = plt.subplots(figsize=(width, height))
  axs.axis('off')
  fig.tight_layout(rect=(0, 0, 1, 1), pad=0.15)
  if title:
    plt.title(title)
  grid = ImageGrid(fig, 111, nrows_ncols=(rows, cols), axes_pad=axes_pad)
  for ax, im in zip(grid, images):
    simple_image_axis(im, ax)
  if save_path:
    plt.savefig(save_path)
  if show:
    plt.show()
  return axs


